@@ -62,9 +62,7 @@ node_new (int vecmax)
 static Binding
 search_node (Binding tree, size_t key)
 {
-  size_t i;
-
-  for (i = 0; i < tree->vecnum; ++i)
+  for (size_t i = 0; i < tree->vecnum; ++i)
     if (tree->vec[i]->key == key)
       return tree->vec[i];
 
@@ -132,32 +130,27 @@ search_key (Binding tree, gl_list_t keys, size_t from)
 size_t
 do_binding_completion (astr as)
 {
-  size_t key;
   astr bs = astr_new ();
 
   if (lastflag & FLAG_SET_UNIARG)
     {
-      int arg = last_uniarg;
-
-      if (arg < 0)
+      unsigned arg = abs (last_uniarg);
+      do
         {
-          astr_cat_cstr (bs, "- ");
-          arg = -arg;
+          bs = astr_fmt ("%c %s", arg % 10 + '0', astr_cstr (bs));
+          arg /= 10;
         }
+      while (arg != 0);
 
-      do {
-        astr_insert_char (bs, 0, ' ');
-        astr_insert_char (bs, 0, arg % 10 + '0');
-        arg /= 10;
-      } while (arg != 0);
+      if (last_uniarg < 0)
+        bs = astr_fmt ("- %s", astr_cstr (bs));
     }
 
   minibuf_write ("%s%s%s-",
                  lastflag & (FLAG_SET_UNIARG | FLAG_UNIARG_EMPTY) ? "C-u " : "",
                  astr_cstr (bs),
                  astr_cstr (as));
-  astr_delete (bs);
-  key = getkey ();
+  size_t key = getkey (GETKEY_DEFAULT);
   minibuf_clear ();
 
   return key;
@@ -173,7 +166,7 @@ get_key_sequence (void)
   size_t key;
 
   do
-    key = getkey ();
+    key = getkey (GETKEY_DEFAULT);
   while (key == KBD_NOKEY);
   gl_list_add_last (keys, (void *) key);
   for (;;)
@@ -184,7 +177,6 @@ get_key_sequence (void)
         break;
       as = keyvectostr (keys);
       gl_list_add_last (keys, (void *) do_binding_completion (as));
-      astr_delete (as);
     }
 
   return keys;
@@ -258,28 +250,15 @@ set_this_command (Function cmd)
   _this_command = cmd;
 }
 
-void
-process_command (void)
+le *
+call_command (Function f, int uniarg, bool uniflag, le *branch)
 {
-  gl_list_t keys = get_key_sequence ();
-  Function f = get_function_by_keys (keys);
-
   thisflag = lastflag & FLAG_DEFINING_MACRO;
-  minibuf_clear ();
 
-  if (f != NULL)
-    {
-      set_this_command (f);
-      f (last_uniarg, (lastflag & FLAG_SET_UNIARG) != 0, NULL);
-      _last_command = _this_command;
-    }
-  else
-    {
-      astr as = keyvectostr (keys);
-      minibuf_error ("%s is undefined", astr_cstr (as));
-      astr_delete (as);
-    }
-  gl_list_free (keys);
+  /* Execute the command. */
+  _this_command = f;
+  le *ok = f (uniarg, uniflag, branch);
+  _last_command = _this_command;
 
   /* Only add keystrokes if we were already in macro defining mode
      before the function call, to cope with start-kbd-macro. */
@@ -289,10 +268,26 @@ process_command (void)
   if (!(thisflag & FLAG_SET_UNIARG))
     last_uniarg = 1;
 
-  if (last_command () != F_undo)
+  if (cur_bp && last_command () != F_undo)
     set_buffer_next_undop (cur_bp, get_buffer_last_undop (cur_bp));
 
   lastflag = thisflag;
+
+  return ok;
+}
+
+void
+process_command (void)
+{
+  gl_list_t keys = get_key_sequence ();
+  Function f = get_function_by_keys (keys);
+
+  minibuf_clear ();
+
+  if (f != NULL)
+    call_command (f, last_uniarg, (lastflag & FLAG_SET_UNIARG) != 0, NULL);
+  else
+    minibuf_error ("%s is undefined", astr_cstr (keyvectostr (keys)));
 }
 
 static Binding
@@ -304,16 +299,13 @@ init_bindings (void)
 void
 init_default_bindings (void)
 {
-  size_t i;
-  gl_list_t keys = gl_list_create_empty (GL_ARRAY_LIST,
-                                         NULL, NULL, NULL, true);
-  astr as;
-
   root_bindings = init_bindings ();
 
   /* Bind all printing keys to self_insert_command */
+  gl_list_t keys = gl_list_create_empty (GL_ARRAY_LIST,
+                                         NULL, NULL, NULL, true);
   gl_list_add_last (keys, NULL);
-  for (i = 0; i <= 0xff; i++)
+  for (size_t i = 0; i <= 0xff; i++)
     {
       if (isprint (i))
         {
@@ -321,9 +313,8 @@ init_default_bindings (void)
           bind_key_vec (root_bindings, keys, 0, F_self_insert_command);
         }
     }
-  gl_list_free (keys);
 
-  as = astr_new_cstr ("\
+  astr as = astr_new_cstr ("\
 (global-set-key \"\\M-m\" 'back-to-indentation)\
 (global-set-key \"\\LEFT\" 'backward-char)\
 (global-set-key \"\\C-b\" 'backward-char)\
@@ -400,7 +391,6 @@ init_default_bindings (void)
 (global-set-key \"\\C-n\" 'next-line)\
 (global-set-key \"\\C-o\" 'open-line)\
 (global-set-key \"\\C-xo\" 'other-window)\
-(global-set-key \"\\INSERT\" 'overwrite-mode)\
 (global-set-key \"\\UP\" 'previous-line)\
 (global-set-key \"\\C-p\" 'previous-line)\
 (global-set-key \"\\M-%\" 'query-replace)\
@@ -433,15 +423,12 @@ init_default_bindings (void)
 (global-set-key \"\\C-u\" 'universal-argument)\
 (global-set-key \"\\C-x\\C-u\" 'upcase-region)\
 (global-set-key \"\\M-u\" 'upcase-word)\
-(global-set-key \"\\C-h\\C-f\" 'view-emacs-FAQ)\
-(global-set-key \"\\F1\\C-f\" 'view-emacs-FAQ)\
 (global-set-key \"\\C-hw\" 'where-is)\
 (global-set-key \"\\F1w\" 'where-is)\
 (global-set-key \"\\C-x\\C-w\" 'write-file)\
 (global-set-key \"\\C-y\" 'yank)\
 ");
   lisp_loadstring (as);
-  astr_delete (as);
 }
 
 DEFUN_ARGS ("global-set-key", global_set_key,
@@ -459,22 +446,18 @@ sequence.
   STR_INIT (keystr);
   if (keystr != NULL)
     {
-      keys = keystrtovec (keystr);
+      keys = keystrtovec (astr_cstr (keystr));
       if (keys == NULL)
         {
-          minibuf_error ("Key sequence %s is invalid", keystr);
+          minibuf_error ("Key sequence %s is invalid", astr_cstr (keystr));
           return leNIL;
         }
     }
   else
     {
-      astr as;
-
       minibuf_write ("Set key globally: ");
       keys = get_key_sequence ();
-      as = keyvectostr (keys);
-      keystr = xstrdup (astr_cstr (as));
-      astr_delete (as);
+      keystr = keyvectostr (keys);
     }
 
   STR_INIT (name)
@@ -484,65 +467,49 @@ sequence.
   if (name == NULL)
     return leNIL;
 
-  func = get_function (name);
+  func = get_function (astr_cstr (name));
   if (func == NULL) /* Possible if called non-interactively */
     {
-      minibuf_error ("No such function `%s'", name);
+      minibuf_error ("No such function `%s'", astr_cstr (name));
       return leNIL;
     }
   bind_key_vec (root_bindings, keys, 0, func);
-
-  gl_list_free (keys);
-  STR_FREE (keystr);
-  STR_FREE (name);
 }
 END_DEFUN
 
-/* FIXME: This function relies on chordtostr (0) returning ""; it
-   should not call chordtostr in this case, should iterate from j=0,
-   not j=1, and should only remove an element from the array at the
-   end of the function if one was added. */
 static void
 walk_bindings_tree (Binding tree, gl_list_t keys,
                     void (*process) (astr key, Binding p, void *st), void *st)
 {
-  size_t i, j;
-  astr as = chordtostr (tree->key);
-
-  gl_list_add_last (keys, as);
-
-  for (i = 0; i < tree->vecnum; ++i)
+  for (size_t i = 0; i < tree->vecnum; ++i)
     {
       Binding p = tree->vec[i];
       if (p->func != NULL)
         {
           astr key = astr_new ();
-          astr bs = chordtostr (p->key);
-          for (j = 1; j < gl_list_size (keys); j++)
+          for (size_t j = 0; j < gl_list_size (keys); j++)
             {
               astr_cat (key, (castr) gl_list_get_at (keys, j));
               astr_cat_char (key, ' ');
             }
-          astr_cat (key, bs);
-          astr_delete (bs);
+          astr_cat (key, chordtostr (p->key));
           process (key, p, st);
-          astr_delete (key);
         }
       else
-        walk_bindings_tree (p, keys, process, st);
+        {
+          gl_list_add_last (keys, chordtostr (p->key));
+          walk_bindings_tree (p, keys, process, st);
+          assert (gl_list_remove_at (keys, gl_list_size (keys) - 1));
+        }
     }
-
-  assert (gl_list_remove_at (keys, gl_list_size (keys) - 1));
 }
 
 static void
 walk_bindings (Binding tree, void (*process) (astr key, Binding p, void *st),
                void *st)
 {
-  gl_list_t l = gl_list_create_empty (GL_LINKED_LIST,
-                                      NULL, NULL, (gl_listelement_dispose_fn) astr_delete, true);
-  walk_bindings_tree (tree, l, process, st);
-  gl_list_free (l);
+  walk_bindings_tree (tree, gl_list_create_empty (GL_LINKED_LIST,
+                                                  NULL, NULL, NULL, true), process, st);
 }
 
 typedef struct
@@ -567,41 +534,29 @@ gather_bindings (astr key, Binding p, void *st)
 DEFUN ("where-is", where_is)
 /*+
 Print message listing key sequences that invoke the command DEFINITION.
-Argument is a command name.  If the prefix arg is non-nil, insert the
-message in the buffer.
+Argument is a command name.
 +*/
 {
-  char *name = minibuf_read_function_name ("Where is command: ");
+  castr name = minibuf_read_function_name ("Where is command: ");
   gather_bindings_state g;
 
   ok = leNIL;
 
   if (name)
     {
-      g.f = get_function (name);
+      g.f = get_function (astr_cstr (name));
       if (g.f)
         {
           g.bindings = astr_new ();
           walk_bindings (root_bindings, gather_bindings, &g);
 
           if (astr_len (g.bindings) == 0)
-            minibuf_write ("%s is not on any key", name);
+            minibuf_write ("%s is not on any key", astr_cstr (name));
           else
-            {
-              astr as = astr_new ();
-              astr_afmt (as, "%s is on %s", name, astr_cstr (g.bindings));
-              if (lastflag & FLAG_SET_UNIARG)
-                bprintf ("%s", astr_cstr (as));
-              else
-                minibuf_write ("%s", astr_cstr (as));
-              astr_delete (as);
-            }
-          astr_delete (g.bindings);
+            minibuf_write ("%s is on %s", astr_cstr (name), astr_cstr (g.bindings));
           ok = leT;
         }
     }
-
-  free (name);
 }
 END_DEFUN
 
