@@ -22,14 +22,10 @@
 #include "gl_xlist.h"
 
 /* basic.c ---------------------------------------------------------------- */
-size_t get_goalc_bp (Buffer * bp, Point pt);
+size_t get_goalc_bp (Buffer * bp, size_t o);
 size_t get_goalc (void);
 bool previous_line (void);
 bool next_line (void);
-bool backward_char (void);
-bool forward_char (void);
-void gotobob (void);
-void gotoeob (void);
 
 /* bind.c ----------------------------------------------------------------- */
 Function last_command (void);
@@ -37,10 +33,15 @@ void set_this_command (Function cmd);
 size_t do_binding_completion (astr as);
 gl_list_t get_key_sequence (void);
 Function get_function_by_keys (gl_list_t keys);
+le *call_command (Function f, int uniarg, bool uniflag, le *branch);
 void process_command (void);
 void init_default_bindings (void);
 
 /* buffer.c --------------------------------------------------------------- */
+int insert_char (int c);
+bool delete_char (void);
+bool replace_estr (size_t del, estr es);
+bool insert_estr (estr as);
 #define FIELD(ty, field)                                \
   ty get_buffer_ ## field (const Buffer *bp);           \
   void set_buffer_ ## field (Buffer *bp, ty field);
@@ -50,33 +51,49 @@ void init_default_bindings (void);
 #include "buffer.h"
 #undef FIELD
 #undef FIELD_STR
-#define FIELD(ty, field)                                \
-  ty get_region_ ## field (const Region *cp);           \
-  void set_region_ ## field (Region *cp, ty field);
-#include "region.h"
-#undef FIELD
+void set_buffer_text (Buffer *bp, estr es);
+castr get_buffer_pre_point (Buffer *bp);
+castr get_buffer_post_point (Buffer *bp);
+size_t get_buffer_pt (Buffer *bp);
+void set_buffer_pt (Buffer *bp, size_t o);
+size_t get_buffer_size (Buffer * bp);
+const char *get_buffer_eol (Buffer *bp);
+size_t buffer_prev_line (Buffer *bp, size_t o);
+size_t buffer_next_line (Buffer *bp, size_t o);
+size_t buffer_start_of_line (Buffer *bp, size_t o);
+size_t buffer_end_of_line (Buffer *bp, size_t o);
+size_t buffer_line_len (Buffer *bp, size_t o);
+size_t get_region_size (const Region r);
+size_t get_buffer_line_o (Buffer *bp);
+char get_buffer_char (Buffer *bp, size_t o);
 void free_buffer (Buffer * bp);
 void init_buffer (Buffer * bp);
+void insert_buffer (Buffer * bp);
 Buffer * buffer_new (void);
 const char *get_buffer_filename_or_name (Buffer * bp);
 void set_buffer_names (Buffer * bp, const char *filename);
 Buffer * find_buffer (const char *name);
 void switch_to_buffer (Buffer * bp);
-Region * region_new (void);
+int warn_if_no_mark (void);
 int warn_if_readonly_buffer (void);
-int calculate_the_region (Region * rp);
-bool delete_region (const Region * rp);
-bool in_region (size_t lineno, size_t x, Region * rp);
+Region region_new (size_t o1, size_t o2);
+Region calculate_the_region (void);
+bool delete_region (const Region r);
+bool in_region (size_t o, size_t x, Region r);
 void set_temporary_buffer (Buffer * bp);
-size_t calculate_buffer_size (Buffer * bp);
 void activate_mark (void);
 void deactivate_mark (void);
 size_t tab_width (Buffer * bp);
-astr copy_text_block (Point pt, size_t size);
+estr get_buffer_region (Buffer *bp, Region r);
+Buffer *create_auto_buffer (const char *name);
 Buffer *create_scratch_buffer (void);
 void kill_buffer (Buffer * kill_bp);
 Completion *make_buffer_completion (void);
 bool check_modified_buffer (Buffer * bp);
+bool move_char (int dir);
+bool move_line (int n);
+size_t offset_to_line (Buffer *bp, size_t offset);
+void goto_offset (size_t o);
 
 /* completion.c ----------------------------------------------------------- */
 #define FIELD(ty, field)                                        \
@@ -90,15 +107,11 @@ bool check_modified_buffer (Buffer * bp);
 #undef FIELD_STR
 int completion_strcmp (const void *p1, const void *p2);
 Completion *completion_new (int fileflag);
-void free_completion (Completion * cp);
 void completion_scroll_up (void);
 void completion_scroll_down (void);
 int completion_try (Completion * cp, astr search, int popup_when_complete);
 
 /* editfns.c -------------------------------------------------------------- */
-void push_mark (void);
-void pop_mark (void);
-void set_mark (void);
 bool is_empty_line (void);
 bool is_blank_line (void);
 int following_char (void);
@@ -115,25 +128,23 @@ size_t countNodes (le * branch);
 void leEval (le * list);
 le *execute_with_uniarg (bool undo, int uniarg, bool (*forward) (void),
                          bool (*backward) (void));
-le *execute_function (const char *name, int uniarg);
+le *move_with_uniarg (int uniarg, bool (*move) (int dir));
+le *execute_function (const char *name, int uniarg, bool is_uniarg);
 Function get_function (const char *name);
 const char *get_function_doc (const char *name);
 int get_function_interactive (const char *name);
 const char *get_function_name (Function p);
-char *minibuf_read_function_name (const char *fmt, ...);
+castr minibuf_read_function_name (const char *fmt, ...);
 void init_eval (void);
 
 /* file.c ----------------------------------------------------------------- */
-extern const char *coding_eol_lf;
-extern const char *coding_eol_crlf;
-extern const char *coding_eol_cr;
 int exist_file (const char *filename);
 astr get_home_dir (void);
 astr agetcwd (void);
 bool expand_path (astr path);
 astr compact_path (astr path);
 bool find_file (const char *filename);
-void zile_exit (int doabort) _GL_ATTRIBUTE_NORETURN;
+_Noreturn void zile_exit (int doabort);
 
 /* funcs.c ---------------------------------------------------------------- */
 void set_mark_interactive (void);
@@ -143,15 +154,14 @@ void write_temp_buffer (const char *name, bool show, void (*func) (va_list ap),
 void pushkey (size_t key);
 void ungetkey (size_t key);
 size_t lastkey (void);
-size_t xgetkey (int mode, size_t timeout);
-size_t getkey (void);
-void waitkey (size_t delay);
+size_t getkey (int delay);
+size_t getkey_unfiltered (int delay);
+void waitkey (void);
 void init_getkey (void);
 void free_getkey (void);
 
 /* history.c -------------------------------------------------------------- */
 History *history_new (void);
-void free_history (History * hp);
 void add_history_element (History * hp, const char *string);
 void prepare_history (History * hp);
 const char *previous_history_element (History * hp);
@@ -163,26 +173,10 @@ gl_list_t keystrtovec (const char *key);
 astr keyvectostr (gl_list_t keys);
 
 /* line.c ----------------------------------------------------------------- */
-#define FIELD(ty, field)                                \
-  ty get_line_ ## field (const Line *wp);               \
-  void set_line_ ## field (Line *wp, ty field);
-  FIELD(const char *, field)
-#include "line.h"
-#undef FIELD
-Line *line_new (void);
-void line_delete (Line *l);
-Line *line_insert (Line *l, astr i);
-void line_remove (Line *l);
-void line_replace_text (Line * lp, size_t offset, size_t oldlen,
-                        const char *newtext, int replace_case);
-int insert_char (int c);
-int insert_char_in_insert_mode (int c);
-bool fill_break_line (void);
 bool insert_newline (void);
-void insert_nstring (const char *s, size_t len);
-void insert_astr (astr as);
+bool intercalate_newline (void);
+bool fill_break_line (void);
 void bprintf (const char *fmt, ...);
-bool delete_char (void);
 
 /* lisp.c ----------------------------------------------------------------- */
 void init_lisp (void);
@@ -194,9 +188,6 @@ void cancel_kbd_macro (void);
 void add_cmd_to_macro (void);
 void add_key_to_cmd (size_t key);
 void remove_key_from_cmd (void);
-void call_macro (Macro * mp);
-Macro *get_macro (const char *name);
-void add_macros_to_list (gl_list_t l, gl_listelement_compar_fn f);
 
 /* main.c ----------------------------------------------------------------- */
 extern char *prog_name;
@@ -212,10 +203,12 @@ extern int thisflag, lastflag, last_uniarg;
 #undef FIELD
 Marker * marker_new (void);
 void unchain_marker (const Marker * marker);
-void free_marker (Marker * marker);
-void move_marker (Marker * marker, Buffer * bp, Point pt);
+void move_marker (Marker * marker, Buffer * bp, size_t o);
 Marker *copy_marker (const Marker * marker);
 Marker *point_marker (void);
+void push_mark (void);
+void pop_mark (void);
+void set_mark (void);
 
 /* minibuf.c -------------------------------------------------------------- */
 void init_minibuf (void);
@@ -223,30 +216,21 @@ int minibuf_no_error (void);
 void minibuf_refresh (void);
 void minibuf_write (const char *fmt, ...);
 void minibuf_error (const char *fmt, ...);
-char *minibuf_read (const char *fmt, const char *value, ...);
-unsigned long minibuf_read_number (const char *fmt, ...);
+castr minibuf_read (const char *fmt, const char *value, ...);
+long minibuf_read_number (const char *fmt, ...);
 bool minibuf_test_in_completions (const char *ms, gl_list_t completions);
 int minibuf_read_yn (const char *fmt, ...);
 int minibuf_read_yesno (const char *fmt, ...);
-char *minibuf_read_completion (const char *fmt, const char *value, Completion * cp,
+castr minibuf_read_completion (const char *fmt, const char *value, Completion * cp,
                                History * hp, ...);
-char *minibuf_vread_completion (const char *fmt, const char *value, Completion * cp,
-                                History * hp, const char *empty_err,
-                                bool (*test) (const char *s, gl_list_t completions),
-                                const char *invalid_err, va_list ap);
-char *minibuf_read_filename (const char *fmt, const char *value,
+castr minibuf_vread_completion (const char *fmt, const char *value, Completion * cp,
+                                      History * hp, const char *empty_err,
+                                      bool (*test) (const char *s, gl_list_t completions),
+                                      const char *invalid_err, va_list ap);
+castr minibuf_read_filename (const char *fmt, const char *value,
                              const char *file, ...);
 void minibuf_clear (void);
 
-/* point.c ---------------------------------------------------------------- */
-Point make_point (size_t lineno, size_t offset);
-int cmp_point (Point pt1, Point pt2);
-Point point_min (void);
-Point point_max (void);
-Point line_beginning_position (int count);
-Point line_end_position (int count);
-void goto_point (Point pt);
-
 /* redisplay.c ------------------------------------------------------------ */
 void resync_redisplay (Window * wp);
 void resize_windows (void);
@@ -263,38 +247,34 @@ void term_move (size_t y, size_t x);
 void term_clrtoeol (void);
 void term_refresh (void);
 void term_clear (void);
-void term_addch (int c);
+void term_addstr (const char *s);
 void term_attrset (size_t attr);
 void term_beep (void);
-size_t term_xgetkey (int mode, size_t timeout);
+size_t term_width (void);
+size_t term_height (void);
+size_t term_getkey (int delay);
+int term_getkey_unfiltered (int delay);
 void term_ungetkey (size_t key);
 
 /* term_minibuf.c --------------------------------------------------------- */
 void term_minibuf_write (const char *fmt);
-char *term_minibuf_read (const char *prompt, const char *value, size_t pos,
+castr term_minibuf_read (const char *prompt, const char *value, size_t pos,
                          Completion * cp, History * hp);
 
 /* term_redisplay.c ------------------------------------------------------- */
-size_t term_width (void);
-size_t term_height (void);
-void term_set_size (size_t cols, size_t rows);
 void term_redisplay (void);
-void term_full_redisplay (void);
-void show_splash_screen (const char *splash);
 void term_finish (void);
-void term_addnstr (const char *s, size_t len);
 
 /* undo.c ----------------------------------------------------------------- */
 extern int undo_nosave;
 void undo_start_sequence (void);
 void undo_end_sequence (void);
-void undo_save (int type, Point pt, size_t arg1, size_t arg2);
-void free_undo (Undo *up);
+void undo_save_block (size_t o, size_t osize, size_t size);
 void undo_set_unchanged (Undo *up);
 
 /* variables.c ------------------------------------------------------------ */
 void init_variables (void);
-char *minibuf_read_variable_name (const char *fmt, ...);
+castr minibuf_read_variable_name (const char *fmt, ...);
 void set_variable (const char *var, const char *val);
 const char *get_variable_doc (const char *var, const char **defval);
 const char *get_variable (const char *var);
@@ -306,7 +286,6 @@ bool get_variable_bool (const char *var);
 #define FIELD(ty, field)                                \
   ty get_window_ ## field (const Window *wp);           \
   void set_window_ ## field (Window *wp, ty field);
-  FIELD(const char *, field)
 #include "window.h"
 #undef FIELD
 void create_scratch_window (void);
@@ -314,7 +293,7 @@ Window *find_window (const char *name);
 Window *popup_window (void);
 void set_current_window (Window * wp);
 void delete_window (Window * del_wp);
-Point window_pt (Window * wp);
+size_t window_o (Window * wp);
 bool window_top_visible (Window * wp);
 bool window_bottom_visible (Window * wp);
 
