@@ -31,6 +31,8 @@
 #include <unistd.h>
 #include <utime.h>
 #include "dirname.h"
+#include "xgetcwd.h"
+#include "copy-file.h"
 
 #include "main.h"
 #include "extern.h"
@@ -39,12 +41,7 @@ int
 exist_file (const char *filename)
 {
   struct stat st;
-
-  if (stat (filename, &st) == -1)
-    if (errno == ENOENT)
-      return false;
-
-  return true;
+  return !(stat (filename, &st) == -1 && errno == ENOENT);
 }
 
 static int
@@ -60,10 +57,8 @@ is_regular_file (const char *filename)
 astr
 agetcwd (void)
 {
-  char *s = getcwd (NULL, 0);
-  astr as = astr_new_cstr (s ? s : "");
-  free (s);
-  return as;
+  char *s = xgetcwd ();
+  return astr_new_cstr (s ? s : "");
 }
 
 /*
@@ -80,20 +75,18 @@ bool
 expand_path (astr path)
 {
   int ok = true;
-  const char *sp = astr_cstr (path), *p;
+  const char *sp = astr_cstr (path);
   astr epath = astr_new ();
 
   if (*sp != '/' && *sp != '~')
     {
-      astr cwd = agetcwd ();
-      astr_cat (epath, cwd);
-      astr_delete (cwd);
+      astr_cat (epath, agetcwd ());
       if (astr_len (epath) == 0 ||
           astr_get (epath, astr_len (epath) - 1) != '/')
         astr_cat_char (epath, '/');
     }
 
-  for (p = sp; *p != '\0';)
+  for (const char *p = sp; *p != '\0';)
     {
       if (*p == '/')
         {
@@ -122,7 +115,7 @@ expand_path (astr path)
                   ok = false;
                   break;
                 }
-              if (STRNEQ (pw->pw_dir, "/"))
+              if (!STREQ (pw->pw_dir, "/"))
                 astr_cat_cstr (epath, pw->pw_dir);
             }
           else
@@ -131,7 +124,6 @@ expand_path (astr path)
               while (*p != '\0' && *p != '/')
                 astr_cat_char (as, *p++);
               pw = getpwnam (astr_cstr (as));
-              astr_delete (as);
               if (pw == NULL)
                 {
                   ok = false;
@@ -159,7 +151,6 @@ expand_path (astr path)
     }
 
   astr_cpy (path, epath);
-  astr_delete (epath);
 
   return ok;
 }
@@ -177,45 +168,19 @@ compact_path (astr path)
     {
       /* Replace `/userhome/' (if found) with `~/'. */
       size_t homelen = strlen (pw->pw_dir);
-      if (astr_len (path) >= homelen &&
-          !strncmp (pw->pw_dir, astr_cstr (path), homelen))
-        {
-          astr buf = astr_new_cstr ("~/");
-          if (STREQ (pw->pw_dir, "/"))
-            astr_cat_cstr (buf, astr_cstr (path) + 1);
-          else
-            astr_cat_cstr (buf, astr_cstr (path) + homelen + 1);
-          astr_cpy (path, buf);
-          astr_delete (buf);
-        }
+      if (homelen > 0 && pw->pw_dir[homelen - 1] == '/')
+        homelen--;
+
+      if (astr_len (path) > homelen &&
+          !strncmp (pw->pw_dir, astr_cstr (path), homelen) &&
+          astr_get (path, homelen) == '/')
+        astr_cpy (path, astr_cat_cstr (astr_new_cstr ("~/"),
+                                       astr_cstr (path) + homelen + 1));
     }
 
   return path;
 }
 
-/*
- * Return the current directory for the buffer, always ending in `/'.
- */
-static astr
-get_buffer_dir (void)
-{
-  astr ret;
-
-  if (get_buffer_filename (cur_bp))
-    {
-      char *name = dir_name (get_buffer_filename (cur_bp));
-      ret = astr_new_cstr (name);
-      free (name);
-    }
-  else
-    ret = agetcwd ();
-
-  if (astr_len (ret) != 0 && astr_get (ret, astr_len (ret) - 1) != '/')
-    astr_cat_char (ret, '/');
-
-  return ret;
-}
-
 /*
  * Get HOME directory.
  */
@@ -228,147 +193,51 @@ get_home_dir (void)
   return NULL;
 }
 
-/* Return nonzero if file exists and can be written. */
-static int
+/* Return true if file exists and can be written. */
+static bool
 check_writable (const char *filename)
 {
   return euidaccess (filename, W_OK) >= 0;
 }
 
-/* Formats of end-of-line. */
-const char *coding_eol_lf = "\n";
-const char *coding_eol_crlf = "\r\n";
-const char *coding_eol_cr = "\r";
-
-/* Maximum number of EOLs to check before deciding type. */
-#define MAX_EOL_CHECK_COUNT 3
-
-/*
- * Read the file contents into current buffer.
- * Return quietly if the file doesn't exist, or other error.
- */
-static void
-read_file (const char *filename)
-{
-  Line *lp;
-  int i, size;
-  bool first_eol = true;
-  const char *this_eol_type;
-  size_t eol_len = 0, total_eols = 0;
-  char buf[BUFSIZ];
-  FILE *fp = fopen (filename, "r");
-  Point pt;
-
-  if (fp == NULL)
-    {
-      if (errno != ENOENT)
-        {
-          minibuf_write ("%s: %s", filename, strerror (errno));
-          set_buffer_readonly (cur_bp, true);
-        }
-      return;
-    }
-
-  if (!check_writable (filename))
-    set_buffer_readonly (cur_bp, true);
-
-  pt = get_buffer_pt (cur_bp);
-  lp = pt.p;
-
-  /* Read first chunk and determine EOL type. */
-  /* FIXME: Don't assume first EOL occurs in first chunk. */
-  size = fread (buf, 1, BUFSIZ, fp);
-  if (size > 0)
-    {
-      for (i = 0; i < size && total_eols < MAX_EOL_CHECK_COUNT; i++)
-        {
-          if (buf[i] == '\n' || buf[i] == '\r')
-            {
-              total_eols++;
-              if (buf[i] == '\n')
-                this_eol_type = coding_eol_lf;
-              else if (i == size - 1 || buf[i + 1] != '\n')
-                this_eol_type = coding_eol_cr;
-              else
-                {
-                  this_eol_type = coding_eol_crlf;
-                  i++;
-                }
-
-              if (first_eol)
-                {
-                  /* This is the first end-of-line. */
-                  set_buffer_eol (cur_bp, this_eol_type);
-                  first_eol = false;
-                }
-              else if (get_buffer_eol (cur_bp) != this_eol_type)
-                {
-                  /* This EOL is different from the last; arbitrarily choose
-                     LF. */
-                  set_buffer_eol (cur_bp, coding_eol_lf);
-                  break;
-                }
-            }
-        }
-
-      /* Process this and subsequent chunks into lines. */
-      eol_len = strlen (get_buffer_eol (cur_bp));
-      do
-        {
-          for (i = 0; i < size; i++)
-            {
-              if (strncmp (get_buffer_eol (cur_bp), buf + i, eol_len) != 0)
-                astr_cat_char (get_line_text (lp), buf[i]);
-              else
-                {
-                  lp = line_insert (lp, astr_new ());
-                  set_buffer_last_line (cur_bp,
-                                        get_buffer_last_line (cur_bp) + 1);
-                  i += eol_len - 1;
-                }
-            }
-        }
-      while ((size = fread (buf, 1, BUFSIZ, fp)) > 0);
-    }
-
-  set_line_next (lp, get_buffer_lines (cur_bp));
-  set_line_prev (get_buffer_lines (cur_bp), lp);
-  pt = get_buffer_pt (cur_bp);
-  pt.p = get_line_next (get_buffer_lines (cur_bp));
-  set_buffer_pt (cur_bp, pt);
-
-  fclose (fp);
-}
-
 bool
 find_file (const char *filename)
 {
   Buffer *bp;
-
   for (bp = head_bp; bp != NULL; bp = get_buffer_next (bp))
+    if (get_buffer_filename (bp) != NULL &&
+        STREQ (get_buffer_filename (bp), filename))
+      break;
+
+  if (bp == NULL)
     {
-      if (get_buffer_filename (bp) != NULL &&
-          STREQ (get_buffer_filename (bp), filename))
+      if (exist_file (filename) && !is_regular_file (filename))
+        minibuf_error ("File exists but could not be read");
+      else
         {
-          switch_to_buffer (bp);
-          return true;
+          bp = buffer_new ();
+          set_buffer_names (bp, filename);
+          set_buffer_dir (bp, astr_new_cstr (dir_name (filename)));
+
+          estr es = estr_readf (filename);
+          if (es.as)
+            set_buffer_readonly (bp, !check_writable (filename));
+          else
+            es = estr_new_astr (astr_new ());
+          set_buffer_text (bp, es);
+
+          /* Reset undo history. */
+          set_buffer_next_undop (bp, NULL);
+          set_buffer_last_undop (bp, NULL);
+          set_buffer_modified (bp, false);
         }
     }
 
-  if (exist_file (filename) && !is_regular_file (filename))
-    {
-      minibuf_error ("File exists but could not be read");
-      return false;
-    }
-
-  bp = buffer_new ();
-  set_buffer_names (bp, filename);
+  if (bp == NULL)
+    return false;
 
   switch_to_buffer (bp);
-  read_file (filename);
-
   thisflag |= FLAG_NEED_RESYNC;
-
   return true;
 }
 
@@ -379,18 +248,15 @@ Switch to a buffer visiting the file,
 creating one if none already exists.
 +*/
 {
-  astr buf = get_buffer_dir ();
-  char *ms = minibuf_read_filename ("Find file: ", astr_cstr (buf), NULL);
+  castr ms = minibuf_read_filename ("Find file: ",
+                                   astr_cstr (get_buffer_dir (cur_bp)), NULL);
 
   ok = leNIL;
 
-  astr_delete (buf);
   if (ms == NULL)
     ok = FUNCALL (keyboard_quit);
-  else if (ms[0] != '\0')
-    ok = bool_to_lisp (find_file (ms));
-
-  free (ms);
+  else if (astr_len (ms) > 0)
+    ok = bool_to_lisp (find_file (astr_cstr (ms)));
 }
 END_DEFUN
 
@@ -416,31 +282,21 @@ If the current buffer now contains an empty file that you just visited
 {
   const char *buf = get_buffer_filename (cur_bp);
   char *base = NULL;
-  char *ms;
-  astr as = NULL;
 
   if (buf == NULL)
-    {
-      as = get_buffer_dir ();
-      buf = astr_cstr (as);
-    }
+    buf = astr_cstr (get_buffer_dir (cur_bp));
   else
     base = base_name (buf);
-  ms = minibuf_read_filename ("Find alternate: ", buf, base);
-  free (base);
+  castr ms = minibuf_read_filename ("Find alternate: ", buf, base);
 
   ok = leNIL;
   if (ms == NULL)
     ok = FUNCALL (keyboard_quit);
-  else if (ms[0] != '\0' && check_modified_buffer (cur_bp))
+  else if (astr_len (ms) > 0 && check_modified_buffer (cur_bp))
     {
       kill_buffer (cur_bp);
-      ok = bool_to_lisp (find_file (ms));
+      ok = bool_to_lisp (find_file (astr_cstr (ms)));
     }
-
-  free (ms);
-  if (as)
-    astr_delete (as);
 }
 END_DEFUN
 
@@ -458,7 +314,6 @@ Select buffer @i{buffer} in the current window.
       Completion *cp = make_buffer_completion ();
       buf = minibuf_read_completion ("Switch to buffer (default %s): ",
                                      "", cp, NULL, get_buffer_name (bp));
-      free_completion (cp);
 
       if (buf == NULL)
         ok = FUNCALL (keyboard_quit);
@@ -466,13 +321,13 @@ Select buffer @i{buffer} in the current window.
 
   if (ok == leT)
     {
-      if (buf && buf[0] != '\0')
+      if (buf && astr_len (buf) > 0)
         {
-          bp = find_buffer (buf);
+          bp = find_buffer (astr_cstr (buf));
           if (bp == NULL)
             {
               bp = buffer_new ();
-              set_buffer_name (bp, buf);
+              set_buffer_name (bp, astr_cstr (buf));
               set_buffer_needname (bp, true);
               set_buffer_nosave (bp, true);
             }
@@ -480,42 +335,9 @@ Select buffer @i{buffer} in the current window.
 
       switch_to_buffer (bp);
     }
-
-  STR_FREE (buf);
 }
 END_DEFUN
 
-static size_t
-insert_lines (size_t n, size_t end, size_t last, Line *from_lp)
-{
-  for (; n < end; n++, from_lp = get_line_next (from_lp))
-    {
-      insert_astr (get_line_text (from_lp));
-      if (n < last)
-        insert_newline ();
-    }
-  return n;
-}
-
-static void
-insert_buffer (Buffer * bp)
-{
-  Line *old_next = get_line_next (get_buffer_pt (bp).p);
-  astr old_cur_line = astr_cpy (astr_new (), get_line_text (get_buffer_pt (bp).p));
-  size_t old_cur_n = get_buffer_pt (bp).n, old_lines = get_buffer_last_line (bp);
-  size_t size = calculate_buffer_size (bp);
-
-  undo_save (UNDO_REPLACE_BLOCK, get_buffer_pt (cur_bp), 0, size);
-  undo_nosave = true;
-  insert_lines (0, old_cur_n, old_lines, get_line_next (get_buffer_lines (bp)));
-  insert_astr (old_cur_line);
-  if (old_cur_n < old_lines)
-    insert_newline ();
-  insert_lines (old_cur_n + 1, old_lines, old_lines, old_next);
-  astr_delete (old_cur_line);
-  undo_nosave = false;
-}
-
 DEFUN_ARGS ("insert-buffer", insert_buffer,
             STR_ARG (buf))
 /*+
@@ -536,71 +358,33 @@ Puts mark after the inserted text.
                                      "", cp, NULL, get_buffer_name (def_bp));
       if (buf == NULL)
         ok = FUNCALL (keyboard_quit);
-      free_completion (cp);
     }
 
   if (ok == leT)
     {
       Buffer *bp;
 
-      if (buf && buf[0] != '\0')
+      if (buf && astr_len (buf) > 0)
         {
-          bp = find_buffer (buf);
+          bp = find_buffer (astr_cstr (buf));
           if (bp == NULL)
             {
-              minibuf_error ("Buffer `%s' not found", buf);
+              minibuf_error ("Buffer `%s' not found", astr_cstr (buf));
               ok = leNIL;
             }
         }
       else
         bp = def_bp;
 
-      insert_buffer (bp);
-      set_mark_interactive ();
+      if (ok != leNIL)
+        {
+          insert_buffer (bp);
+          FUNCALL (set_mark_command);
+        }
     }
-
-  STR_FREE (buf);
 }
 END_DEFUN
 
-static int
-insert_file (const char *filename)
-{
-  int fd;
-  size_t size;
-  char buf[BUFSIZ];
-
-  if (!exist_file (filename))
-    {
-      minibuf_error ("Unable to read file `%s'", filename);
-      return false;
-    }
-
-  fd = open (filename, O_RDONLY);
-  if (fd < 0)
-    {
-      minibuf_write ("%s: %s", filename, strerror (errno));
-      return false;
-    }
-
-  size = lseek (fd, 0, SEEK_END);
-  if (size < 1)
-    {
-      close (fd);
-      return true;
-    }
-
-  lseek (fd, 0, SEEK_SET);
-  undo_save (UNDO_REPLACE_BLOCK, get_buffer_pt (cur_bp), 0, size);
-  undo_nosave = true;
-  while ((size = read (fd, buf, BUFSIZ)) > 0)
-    insert_nstring (buf, size);
-  undo_nosave = false;
-  close (fd);
-
-  return true;
-}
-
 DEFUN_ARGS ("insert-file", insert_file,
        STR_ARG (file))
 /*+
@@ -614,143 +398,52 @@ Set mark after the inserted text.
   STR_INIT (file)
   else
     {
-      astr buf = get_buffer_dir ();
-      file = minibuf_read_filename ("Insert file: ", astr_cstr (buf), NULL);
+      file = minibuf_read_filename ("Insert file: ",
+                                    astr_cstr (get_buffer_dir (cur_bp)), NULL);
       if (file == NULL)
         ok = FUNCALL (keyboard_quit);
     }
 
-  if (file == NULL || file[0] == '\0' || !insert_file (file))
+  if (file == NULL || astr_len (file) == 0)
     ok = leNIL;
-  else
-    set_mark_interactive ();
 
-  STR_FREE (file);
+  if (ok != leNIL)
+    {
+      estr es = estr_readf (astr_cstr (file));
+      if (es.as != NULL)
+        insert_estr (es);
+      else
+        {
+          ok = leNIL;
+          minibuf_error ("%s: %s", astr_cstr (file), strerror (errno));
+        }
+    }
+  else
+    FUNCALL (set_mark_command);
 }
 END_DEFUN
 
-/*
- * Copy a file.
- */
-static int
-copy_file (const char *source, const char *dest)
-{
-  char buf[BUFSIZ];
-  int ifd, ofd, stat_valid, serrno;
-  size_t len;
-  struct stat st;
-  char *tname;
-
-  ifd = open (source, O_RDONLY, 0);
-  if (ifd < 0)
-    {
-      minibuf_error ("%s: unable to backup", source);
-      return false;
-    }
-
-  tname = xasprintf ("%s_XXXXXX", dest);
-  ofd = mkstemp (tname);
-  if (ofd == -1)
-    {
-      serrno = errno;
-      close (ifd);
-      minibuf_error ("%s: unable to create backup", dest);
-      free (tname);
-      errno = serrno;
-      return false;
-    }
-
-  while ((len = read (ifd, buf, sizeof buf)) > 0)
-    if (write (ofd, buf, len) < 0)
-      {
-        minibuf_error ("Unable to write to backup file `%s'", dest);
-        close (ifd);
-        close (ofd);
-        return false;
-      }
-
-  serrno = errno;
-  stat_valid = fstat (ifd, &st) != -1;
-
-  /* Recover file permissions and ownership. */
-  if (stat_valid)
-    {
-#ifdef HAVE_FCHMOD
-      fchmod (ofd, st.st_mode);
-#endif
-      if (fchown (ofd, st.st_uid, st.st_gid))
-        {
-          /* Avoid compiler warning for ignoring return value. */
-        }
-    }
-
-  close (ifd);
-  close (ofd);
-
-  if (stat_valid)
-    {
-      if (rename (tname, dest) == -1)
-        {
-          minibuf_error ("Cannot rename temporary file `%s'",
-                         strerror (errno));
-          unlink (tname);
-          stat_valid = false;
-        }
-    }
-  else if (unlink (tname) == -1)
-    minibuf_error ("Cannot remove temporary file `%s'", strerror (errno));
-
-  free (tname);
-  errno = serrno;
-
-  /* Recover file modification time. */
-  if (stat_valid)
-    {
-      struct utimbuf t;
-      t.actime = st.st_atime;
-      t.modtime = st.st_mtime;
-      utime (dest, &t);
-    }
-
-  return stat_valid;
-}
-
 /*
  * Write buffer to given file name with given mode.
  */
 static int
-raw_write_to_disk (Buffer * bp, const char *filename, mode_t mode)
+write_to_disk (Buffer * bp, const char *filename, mode_t mode)
 {
-  ssize_t eol_len = (ssize_t) strlen (get_buffer_eol (bp)), written;
-  Line *lp;
-  int ret = 0;
   int fd = creat (filename, mode);
-
   if (fd < 0)
     return -1;
 
-  /* Save the lines. */
-  for (lp = get_line_next (get_buffer_lines (bp));
-       lp != get_buffer_lines (bp);
-       lp = get_line_next (lp))
+  int ret = 0;
+  castr as = get_buffer_pre_point (bp);
+  ssize_t written = write (fd, astr_cstr (as), astr_len (as));
+  if (written < 0 || (size_t) written != astr_len (as))
+    ret = written;
+  else
     {
-      ssize_t len = (ssize_t) astr_len (get_line_text (lp));
-
-      written = write (fd, astr_cstr (get_line_text (lp)), len);
-      if (written != len)
-        {
-          ret = written;
-          break;
-        }
-      if (get_line_next (lp) != get_buffer_lines (bp))
-        {
-          written = write (fd, get_buffer_eol (bp), eol_len);
-          if (written != eol_len)
-            {
-              ret = written;
-              break;
-            }
-        }
+      as = get_buffer_post_point (bp);
+      written = write (fd, astr_cstr (as), astr_len (as));
+      if (written < 0 || (size_t) written != astr_len (as))
+        ret = written;
     }
 
   if (close (fd) < 0 && ret == 0)
@@ -770,25 +463,17 @@ create_backup_filename (const char *filename, const char *backupdir)
   /* Prepend the backup directory path to the filename */
   if (backupdir)
     {
-      astr buf = astr_new ();
-
-      astr_cpy_cstr (buf, backupdir);
+      astr buf = astr_new_cstr (backupdir);
       if (astr_get (buf, astr_len (buf) - 1) != '/')
         astr_cat_char (buf, '/');
-      while (*filename)
-        {
-          if (*filename == '/')
-            astr_cat_char (buf, '!');
-          else
-            astr_cat_char (buf, *filename);
-          ++filename;
-        }
+      for (; *filename; filename++)
+        if (*filename == '/')
+          astr_cat_char (buf, '!');
+        else
+          astr_cat_char (buf, *filename);
 
       if (!expand_path (buf))
-        {
-          astr_delete (buf);
-          buf = NULL;
-        }
+        buf = NULL;
       res = buf;
     }
   else
@@ -802,41 +487,37 @@ create_backup_filename (const char *filename, const char *backupdir)
  * Create a backup file if specified by the user variables.
  */
 static int
-write_to_disk (Buffer * bp, const char *filename)
+backup_and_write (Buffer * bp, const char *filename)
 {
-  int fd, backup = get_variable_bool ("make-backup-files"), ret;
-  const char *backupdir = get_variable_bool ("backup-directory") ?
-    get_variable ("backup-directory") : NULL;
-
   /* Make backup of original file. */
+  int fd, backup = get_variable_bool ("make-backup-files");
   if (!get_buffer_backup (bp) && backup
       && (fd = open (filename, O_RDWR, 0)) != -1)
     {
-      astr bfilename;
       close (fd);
-      bfilename = create_backup_filename (filename, backupdir);
-      if (bfilename && copy_file (filename, astr_cstr (bfilename)))
+
+      const char *backupdir = get_variable_bool ("backup-directory") ?
+        get_variable ("backup-directory") : NULL;
+      astr bfilename = create_backup_filename (filename, backupdir);
+      if (bfilename && copy_file_preserving_error (filename, astr_cstr (bfilename)) == 0)
         set_buffer_backup (bp, true);
       else
         {
           minibuf_error ("Cannot make backup file: %s", strerror (errno));
-          waitkey (WAITKEY_DEFAULT);
+          waitkey ();
         }
-      astr_delete (bfilename);
     }
 
-  ret = raw_write_to_disk (bp, filename, S_IRUSR | S_IWUSR |
+  int ret = write_to_disk (bp, filename, S_IRUSR | S_IWUSR |
                            S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
-  if (ret != 0)
-    {
-      if (ret == -1)
-        minibuf_error ("Error writing `%s': %s", filename, strerror (errno));
-      else
-        minibuf_error ("Error writing `%s'", filename);
-      return false;
-    }
+  if (ret == 0)
+    return true;
 
-  return true;
+  if (ret == -1)
+    minibuf_error ("Error writing `%s': %s", filename, strerror (errno));
+  else
+    minibuf_error ("Error writing `%s'", filename);
+  return false;
 }
 
 static le *
@@ -845,27 +526,24 @@ write_buffer (Buffer *bp, bool needname, bool confirm,
 {
   bool ans = true;
   le * ok = leT;
-  char *name;
+  castr name;
 
   if (!needname)
-    name = xstrdup (name0);
+    name = astr_new_cstr (name0);
 
   if (needname)
     {
       name = minibuf_read_filename (prompt, "", NULL);
       if (name == NULL)
         return FUNCALL (keyboard_quit);
-      if (name[0] == '\0')
-        {
-          free (name);
-          return leNIL;
-        }
+      if (astr_len (name) == 0)
+        return leNIL;
       confirm = true;
     }
 
-  if (confirm && exist_file (name))
+  if (confirm && exist_file (astr_cstr (name)))
     {
-      ans = minibuf_read_yn ("File `%s' exists; overwrite? (y or n) ", name);
+      ans = minibuf_read_yn ("File `%s' exists; overwrite? (y or n) ", astr_cstr (name));
       if (ans == -1)
         FUNCALL (keyboard_quit);
       else if (ans == false)
@@ -876,14 +554,15 @@ write_buffer (Buffer *bp, bool needname, bool confirm,
 
   if (ans == true)
     {
-      if (name != get_buffer_filename (bp))
-        set_buffer_names (bp, name);
+      if (get_buffer_filename (bp) == NULL ||
+          !STREQ (astr_cstr (name), get_buffer_filename (bp)))
+        set_buffer_names (bp, astr_cstr (name));
       set_buffer_needname (bp, false);
       set_buffer_temporary (bp, false);
       set_buffer_nosave (bp, false);
-      if (write_to_disk (bp, name))
+      if (backup_and_write (bp, astr_cstr (name)))
         {
-          minibuf_write ("Wrote %s", name);
+          minibuf_write ("Wrote %s", astr_cstr (name));
           set_buffer_modified (bp, false);
           undo_set_unchanged (get_buffer_last_undop (bp));
         }
@@ -891,26 +570,23 @@ write_buffer (Buffer *bp, bool needname, bool confirm,
         ok = leNIL;
     }
 
-  free (name);
   return ok;
 }
 
 static le *
 save_buffer (Buffer * bp)
 {
-  if (!get_buffer_modified (bp))
-    {
-      minibuf_write ("(No changes need to be saved)");
-      return leT;
-    }
-  else
+  if (get_buffer_modified (bp))
     return write_buffer (bp, get_buffer_needname (bp), false, get_buffer_filename (bp),
                          "File to save in: ");
+
+  minibuf_write ("(No changes need to be saved)");
+  return leT;
 }
 
 DEFUN ("save-buffer", save_buffer)
 /*+
-Save current buffer in visited file if modified. By default, makes the
+Save current buffer in visited file if modified.  By default, makes the
 previous version into a backup file if this is the first save.
 +*/
 {
@@ -935,11 +611,10 @@ END_DEFUN
 static int
 save_some_buffers (void)
 {
-  Buffer *bp;
   bool none_to_save = true;
   bool noask = false;
 
-  for (bp = head_bp; bp != NULL; bp = get_buffer_next (bp))
+  for (Buffer *bp = head_bp; bp != NULL; bp = get_buffer_next (bp))
     {
       if (get_buffer_modified (bp) && !get_buffer_nosave (bp))
         {
@@ -955,7 +630,7 @@ save_some_buffers (void)
                 int c;
 
                 minibuf_write ("Save file %s? (y, n, !, ., q) ", fname);
-                c = getkey ();
+                c = getkey (GETKEY_DEFAULT);
                 minibuf_clear ();
 
                 switch (c)
@@ -981,7 +656,7 @@ save_some_buffers (void)
                     goto exitloop;
                   default:
                     minibuf_error ("Please answer y, n, !, . or q.");
-                    waitkey (WAITKEY_DEFAULT);
+                    waitkey ();
                     break;
                   }
               }
@@ -1012,12 +687,10 @@ DEFUN ("save-buffers-kill-emacs", save_buffers_kill_emacs)
 Offer to save each buffer, then kill this Zile process.
 +*/
 {
-  Buffer *bp;
-
   if (!save_some_buffers ())
     return leNIL;
 
-  for (bp = head_bp; bp != NULL; bp = get_buffer_next (bp))
+  for (Buffer *bp = head_bp; bp != NULL; bp = get_buffer_next (bp))
     if (get_buffer_modified (bp) && !get_buffer_needname (bp))
       {
         for (;;)
@@ -1046,22 +719,15 @@ END_DEFUN
 void
 zile_exit (int doabort)
 {
-  Buffer *bp;
-
   fprintf (stderr, "Trying to save modified buffers (if any)...\r\n");
-  for (bp = head_bp; bp != NULL; bp = get_buffer_next (bp))
+  for (Buffer *bp = head_bp; bp != NULL; bp = get_buffer_next (bp))
     if (get_buffer_modified (bp) && !get_buffer_nosave (bp))
       {
-        astr buf = astr_new (), as;
-        const char *fname = get_buffer_filename_or_name (bp);
-        astr_cpy_cstr (buf, fname);
-        as = astr_new_cstr (PACKAGE);
-        astr_recase (as, case_upper);
-        astr_afmt (buf, ".%sSAVE", astr_cstr (as));
-        astr_delete (as);
+        astr buf = astr_fmt ("%s.%sSAVE",
+                             get_buffer_filename_or_name (bp),
+                             astr_cstr (astr_recase (astr_new_cstr (PACKAGE), case_upper)));
         fprintf (stderr, "Saving %s...\r\n", astr_cstr (buf));
-        raw_write_to_disk (bp, astr_cstr (buf), S_IRUSR | S_IWUSR);
-        astr_delete (buf);
+        write_to_disk (bp, astr_cstr (buf), S_IRUSR | S_IWUSR);
       }
 
   if (doabort)
@@ -1070,43 +736,32 @@ zile_exit (int doabort)
     exit (EXIT_CRASH);
 }
 
-DEFUN ("cd", cd)
+DEFUN_ARGS ("cd", cd,
+       STR_ARG (dir))
 /*+
-Make the user specified directory become the current buffer's default
-directory.
+Make DIR become the current buffer's default directory.
 +*/
 {
-  struct stat st;
-  astr buf = get_buffer_dir ();
-  char *ms = minibuf_read_filename ("Change default directory: ",
-                                    astr_cstr (buf), NULL);
+  if (arglist == NULL)
+    dir = minibuf_read_filename ("Change default directory: ",
+                                 astr_cstr (get_buffer_dir (cur_bp)), NULL);
 
-  if (ms == NULL)
+  if (dir == NULL)
+    ok = FUNCALL (keyboard_quit);
+  else if (astr_len (dir) > 0)
     {
-      astr_delete (buf);
-      return FUNCALL (keyboard_quit);
-    }
-  astr_delete (buf);
+      struct stat st;
 
-  if (ms[0] != '\0')
-    {
-      if (stat (ms, &st) != 0 || !S_ISDIR (st.st_mode))
+      if (stat (astr_cstr (dir), &st) != 0 || !S_ISDIR (st.st_mode))
         {
-          minibuf_error ("`%s' is not a directory", ms);
-          free (ms);
-          return leNIL;
+          minibuf_error ("`%s' is not a directory", astr_cstr (dir));
+          ok = leNIL;
         }
-      if (chdir (ms) == -1)
+      else if (chdir (astr_cstr (dir)) == -1)
         {
-          minibuf_write ("%s: %s", ms, strerror (errno));
-          free (ms);
-          return leNIL;
+          minibuf_write ("%s: %s", dir, strerror (errno));
+          ok = leNIL;
         }
-      free (ms);
-      return leT;
     }
-
-  free (ms);
-  return leNIL;
 }
 END_DEFUN
