@@ -1,6 +1,6 @@
 /* Self documentation facility functions
 
-   Copyright (c) 1997-2004, 2008-2010, 2011 Free Software Foundation, Inc.
+   Copyright (c) 1997-2004, 2008-2011 Free Software Foundation, Inc.
 
    This file is part of GNU Zile.
 
@@ -31,42 +31,6 @@
 #include "main.h"
 #include "extern.h"
 
-DEFUN ("zile-version", zile_version)
-/*+
-Show the version of Zile that is running.
-+*/
-{
-  minibuf_write (ZILE_VERSION_STRING " of " CONFIGURE_DATE " on " CONFIGURE_HOST);
-}
-END_DEFUN
-
-static le *
-show_file (const char *filename)
-{
-  if (!exist_file (filename))
-    {
-      minibuf_error ("Unable to read file `%s'", filename);
-      return leNIL;
-    }
-
-  find_file (filename);
-  set_buffer_readonly (cur_bp, true);
-  set_buffer_noundo (cur_bp, true);
-  set_buffer_needname (cur_bp, true);
-  set_buffer_nosave (cur_bp, true);
-
-  return leT;
-}
-
-DEFUN ("view-emacs-FAQ", view_emacs_FAQ)
-/*+
-Display the Zile Frequently Asked Questions (FAQ) file.
-+*/
-{
-  ok = show_file (PATH_DATA "/FAQ");
-}
-END_DEFUN
-
 static void
 write_function_description (va_list ap)
 {
@@ -86,27 +50,23 @@ DEFUN_ARGS ("describe-function", describe_function,
 Display the full documentation of a function.
 +*/
 {
-  const char *doc;
-
   STR_INIT (func)
   else
     {
-      func = xstrdup (minibuf_read_function_name ("Describe function: "));
+      func = minibuf_read_function_name ("Describe function: ");
       if (func == NULL)
         ok = leNIL;
     }
 
   if (ok == leT)
     {
-      doc = get_function_doc (func);
+      const char *doc = get_function_doc (astr_cstr (func));
       if (doc == NULL)
         ok = leNIL;
       else
         write_temp_buffer ("*Help*", true,
-                           write_function_description, func, doc);
+                           write_function_description, astr_cstr (func), doc);
     }
-
-  STR_FREE (func);
 }
 END_DEFUN
 
@@ -136,16 +96,14 @@ Display the full documentation of a variable.
   else
     {
       const char *defval;
-      const char *doc = get_variable_doc (name, &defval);
+      const char *doc = get_variable_doc (astr_cstr (name), &defval);
       if (doc == NULL)
         ok = leNIL;
       else
         write_temp_buffer ("*Help*", true,
                            write_variable_description,
-                           name, get_variable (name), doc);
+                           astr_cstr (name), get_variable (astr_cstr (name)), doc);
     }
-
-  STR_FREE (name);
 }
 END_DEFUN
 
@@ -178,13 +136,12 @@ Display documentation of the command invoked by a key sequence.
   STR_INIT (keystr);
   if (keystr != NULL)
     {
-      gl_list_t keys = keystrtovec (keystr);
+      gl_list_t keys = keystrtovec (astr_cstr (keystr));
 
       if (keys != NULL)
         {
           name = get_function_name (get_function_by_keys (keys));
           binding = keyvectostr (keys);
-          gl_list_free (keys);
         }
       else
         ok = leNIL;
@@ -197,7 +154,6 @@ Display documentation of the command invoked by a key sequence.
       keys = get_key_sequence ();
       name = get_function_name (get_function_by_keys (keys));
       binding = keyvectostr (keys);
-      gl_list_free (keys);
 
       if (name == NULL)
         {
@@ -217,10 +173,5 @@ Display documentation of the command invoked by a key sequence.
         write_temp_buffer ("*Help*", true,
                            write_key_description, name, doc, astr_cstr (binding));
     }
-
-  if (binding)
-    astr_delete (binding);
-
-  STR_FREE (keystr);
 }
 END_DEFUN
