@@ -1,6 +1,6 @@
 /* Key encoding and decoding functions
 
-   Copyright (c) 1997-2004, 2008-2010, 2011 Free Software Foundation, Inc.
+   Copyright (c) 1997-2004, 2008-2011 Free Software Foundation, Inc.
 
    This file is part of GNU Zile.
 
@@ -128,7 +128,7 @@ chordtostr (size_t key)
       if (key <= 0xff && isgraph (key))
         astr_cat_char (as, (int) key);
       else
-        astr_afmt (as, "<%x>", key);
+        astr_cat (as, astr_fmt ("<%x>", key));
     }
 
   return as;
@@ -142,6 +142,7 @@ static const char *keyname[] = {
   "\\C-",
   "\\DELETE",
   "\\DOWN",
+  "\\e", /* FIXME: Kludge to make keystrings work in both Emacs and Zile. */
   "\\END",
   "\\F1",
   "\\F10",
@@ -167,6 +168,7 @@ static const char *keyname[] = {
   "\\RET",
   "\\RIGHT",
   "\\SPC",
+  "\\t",
   "\\TAB",
   "\\UP",
   "\\\\",
@@ -180,6 +182,7 @@ static int keycode[] = {
   KBD_CTRL,
   KBD_DEL,
   KBD_DOWN,
+  '\033',
   KBD_END,
   KBD_F1,
   KBD_F10,
@@ -206,6 +209,7 @@ static int keycode[] = {
   KBD_RIGHT,
   ' ',
   KBD_TAB,
+  KBD_TAB,
   KBD_UP,
   '\\',
 };
@@ -218,9 +222,8 @@ strtokey (const char *buf, size_t * len)
 {
   if (*buf == '\\')
     {
-      size_t i;
       const char **p = NULL;
-      for (i = 0; i < sizeof (keyname) / sizeof (keyname[0]); i++)
+      for (size_t i = 0; i < sizeof (keyname) / sizeof (keyname[0]); i++)
         if (strncmp (keyname[i], buf, strlen (keyname[i])) == 0)
           p = (const char **) &keyname[i];
       if (p == NULL)
@@ -282,10 +285,7 @@ keystrtovec (const char *key)
     {
       size_t len, code = strtochord (key, &len);
       if (code == KBD_NOKEY)
-        {
-          gl_list_free (keys);
-          return NULL;
-        }
+        return NULL;
       gl_list_add_last (keys, (void *) code);
       key += len;
     }
@@ -299,14 +299,12 @@ keystrtovec (const char *key)
 astr
 keyvectostr (gl_list_t keys)
 {
-  size_t i;
   astr as = astr_new ();
 
-  for (i = 0; i < gl_list_size (keys); i++)
+  for (size_t i = 0; i < gl_list_size (keys); i++)
     {
       astr key = chordtostr ((size_t) gl_list_get_at (keys, i));
       astr_cat (as, key);
-      astr_delete (key);
       if (i < gl_list_size (keys) - 1)
         astr_cat_char (as, ' ');
     }
