@@ -49,64 +49,26 @@ int thisflag = 0, lastflag = 0;
 /* The universal argument repeat count. */
 int last_uniarg = 1;
 
-static char about_splash_str[] = "\
-" ZILE_VERSION_STRING "\n\
+static const char splash_str[] = "\
+Welcome to GNU " PACKAGE_NAME ".\n\
 \n\
+Undo changes       C-x u        Exit " PACKAGE_NAME "         C-x C-c\n\
+(`C-' means use the CTRL key.  `M-' means hold the Meta (or Alt) key.\n\
+If you have no Meta key, you may type ESC followed by the character.)\n\
+Combinations like `C-x u' mean first press `C-x', then `u'.\n\
+Keys not working properly?  See file://" PATH_DATA "/FAQ\n\
+\n\
+" ZILE_VERSION_STRING "\n\
 " ZILE_COPYRIGHT_STRING "\n\
 \n\
-Type `C-x C-c' to exit " PACKAGE_NAME ".\n\
-Type `C-x u' to undo changes.\n\
-Type `C-g' at any time to quit the current operation.\n\
-\n\
-`C-x' means hold the CTRL key while typing the character `x'.\n\
-`M-x' means hold the META or ALT key down while typing `x'.\n\
-If there is no META or ALT key, instead press and release\n\
-the ESC key and then type `x'.\n\
-Combinations like `C-x u' mean first press `C-x', then `u'.\n\
+GNU " PACKAGE_NAME " Zile comes with ABSOLUTELY NO WARRANTY.\n\
+" PACKAGE_NAME " is Free Software--Free as in Freedom--so you can redistribute copies\n\
+of " PACKAGE_NAME " and modify it; see the file COPYING.  Otherwise, a copy can be\n\
+downloaded from http://www.gnu.org/licenses/gpl.html.\n\
 ";
 
-static char about_minibuf_str[] = "Welcome to " PACKAGE_NAME "!";
-
-static void
-about_screen (void)
-{
-  minibuf_write (about_minibuf_str);
-  if (!get_variable_bool ("inhibit-splash-screen"))
-    {
-      show_splash_screen (about_splash_str);
-      term_refresh ();
-      waitkey (20 * 10);
-    }
-}
-
-static void
-setup_main_screen (void)
-{
-  Buffer *bp, *last_bp = NULL;
-  int c = 0;
-
-  for (bp = head_bp; bp; bp = get_buffer_next (bp))
-    {
-      /* Last buffer that isn't *scratch*. */
-      if (get_buffer_next (bp) && !get_buffer_next (get_buffer_next (bp)))
-        last_bp = bp;
-      c++;
-    }
-
-  /* *scratch* and two files. */
-  if (c == 3)
-    {
-      FUNCALL (split_window);
-      switch_to_buffer (last_bp);
-      FUNCALL (other_window);
-    }
-  /* More than two files. */
-  else if (c > 3)
-    FUNCALL (list_buffers);
-}
-
-static void
-_GL_ATTRIBUTE_NORETURN segv_sig_handler (int signo _GL_UNUSED_PARAMETER)
+static void _Noreturn
+segv_sig_handler (int signo _GL_UNUSED_PARAMETER)
 {
   fprintf (stderr,
            "%s: " PACKAGE_NAME
@@ -116,8 +78,8 @@ _GL_ATTRIBUTE_NORETURN segv_sig_handler (int signo _GL_UNUSED_PARAMETER)
   zile_exit (true);
 }
 
-static void
-_GL_ATTRIBUTE_NORETURN other_sig_handler (int signo _GL_UNUSED_PARAMETER)
+static void _Noreturn
+other_sig_handler (int signo _GL_UNUSED_PARAMETER)
 {
   fprintf (stderr, "%s: terminated with signal %d.\r\n",
            program_name, signo);
@@ -160,11 +122,9 @@ main (int argc, char **argv)
                                              NULL, NULL, NULL, false);
   gl_list_t arg_line = gl_list_create_empty (GL_LINKED_LIST,
                                              NULL, NULL, NULL, false);
-  size_t i, line = 1;
-  Buffer *scratch_bp;
-  astr as;
-  bool ok = true;
+  size_t line = 1;
 
+  GC_INIT ();
   set_program_name (argv[0]);
 
   /* Set up Lisp environment now so it's available to files and
@@ -210,14 +170,18 @@ main (int argc, char **argv)
           break;
         case 1:
           gl_list_add_last (arg_type, (void *) ARG_FUNCTION);
-          gl_list_add_last (arg_arg, (void *) optarg);
+          gl_list_add_last (arg_arg, (const void *) astr_new_cstr (optarg));
           gl_list_add_last (arg_line, (void *) 0);
           break;
         case 2:
-          gl_list_add_last (arg_type, (void *) ARG_LOADFILE);
-          gl_list_add_last (arg_arg, (void *) optarg);
-          gl_list_add_last (arg_line, (void *) 0);
-          break;
+          {
+            gl_list_add_last (arg_type, (void *) ARG_LOADFILE);
+            astr as = astr_new_cstr (optarg);
+            expand_path (as);
+            gl_list_add_last (arg_arg, (const void *) astr_cstr (as));
+            gl_list_add_last (arg_line, (void *) 0);
+            break;
+          }
         case 3:
           printf ("Usage: %s [OPTION-OR-FILENAME]...\n"
                   "\n"
@@ -229,9 +193,7 @@ main (int argc, char **argv)
 #define O(longname, shortname, arg, argstring, docstring)               \
           shortopt = xasprintf (", -%c", shortname);                    \
           buf = xasprintf ("--%s%s %s", longname, shortname ? shortopt : "", argstring); \
-          printf ("%-24s%s\n", buf, docstring);                         \
-          free (buf);                                                   \
-          free (shortopt);
+          printf ("%-24s%s\n", buf, docstring);
 #define A(argstring, docstring) \
           printf ("%-24s%s\n", argstring, docstring);
 #include "tbl_opts.h"
@@ -255,7 +217,9 @@ main (int argc, char **argv)
           else
             {
               gl_list_add_last (arg_type, (void *) ARG_FILE);
-              gl_list_add_last (arg_arg, (void *) optarg);
+              astr as = astr_new_cstr (optarg);
+              expand_path (as);
+              gl_list_add_last (arg_arg, (const void *) astr_cstr (as));
               gl_list_add_last (arg_line, (void *) line);
               line = 1;
             }
@@ -277,47 +241,51 @@ main (int argc, char **argv)
   /* Create the `*scratch*' buffer, so that initialisation commands
      that act on a buffer have something to act on. */
   create_scratch_window ();
-  scratch_bp = cur_bp;
-  as = astr_new_cstr ("\
+  Buffer *scratch_bp = cur_bp;
+  bprintf ("%s", "\
 ;; This buffer is for notes you don't want to save.\n\
 ;; If you want to create a file, visit that file with C-x C-f,\n\
 ;; then enter the text in that file's own buffer.\n\
 \n");
-  insert_astr (as);
-  astr_delete (as);
   set_buffer_modified (cur_bp, false);
 
   if (!qflag)
     {
-      as = get_home_dir ();
+      astr as = get_home_dir ();
       if (as)
-        {
-          astr_cat_cstr (as, "/." PACKAGE);
-          lisp_loadfile (astr_cstr (as));
-          astr_delete (as);
-        }
+        lisp_loadfile (astr_cstr (astr_cat_cstr (as, "/." PACKAGE)));
     }
 
-  /* Show the splash screen only if no files, function or load file is
-     specified on the command line, and there has been no error. */
-  if (gl_list_size (arg_arg) == 0 && minibuf_no_error ())
-    about_screen ();
+  /* Create the splash buffer & message only if no files, function or
+     load file is specified on the command line, and there has been no
+     error. */
+  if (gl_list_size (arg_arg) == 0 && minibuf_no_error () &&
+      !get_variable_bool ("inhibit-splash-screen"))
+    {
+      Buffer *bp = create_auto_buffer ("*GNU " PACKAGE_NAME "*");
+      switch_to_buffer (bp);
+      bprintf ("%s", splash_str);
+      set_buffer_readonly (bp, true);
+      FUNCALL (beginning_of_buffer);
+    }
 
   /* Load files and load files and run functions given on the command
      line. */
-  for (i = 0; ok && i < gl_list_size (arg_arg); i++)
+  bool ok = true;
+  for (size_t i = 0; ok && i < gl_list_size (arg_arg); i++)
     {
       const char *arg = (const char *) gl_list_get_at (arg_arg, i);
 
       switch ((ptrdiff_t) gl_list_get_at (arg_type, i))
         {
         case ARG_FUNCTION:
-          ok = get_function (arg) != NULL;
-          if (ok)
-            ok = execute_function (arg, 1) != leNIL;
-          else
-            minibuf_error ("Function `%s' not defined", arg);
-          break;
+          {
+            le *res = execute_function (arg, 1, false);
+            if (res == NULL)
+              minibuf_error ("Function `%s' not defined", arg);
+            ok = res == leT;
+            break;
+          }
         case ARG_LOADFILE:
           ok = lisp_loadfile (arg);
           if (!ok)
@@ -328,7 +296,6 @@ main (int argc, char **argv)
             ok = find_file (arg);
             if (ok)
               FUNCALL_ARG (goto_line, (size_t) gl_list_get_at (arg_line, i));
-            lastflag |= FLAG_NEED_RESYNC;
           }
           break;
         default:
@@ -337,13 +304,26 @@ main (int argc, char **argv)
       if (thisflag & FLAG_QUIT)
         break;
     }
-  gl_list_free (arg_type);
-  gl_list_free (arg_arg);
-  gl_list_free (arg_line);
   lastflag |= FLAG_NEED_RESYNC;
 
   /* Set up screen according to number of files loaded. */
-  setup_main_screen ();
+  Buffer *last_bp = NULL;
+  int c = 0;
+  for (Buffer *bp = head_bp; bp; bp = get_buffer_next (bp))
+    {
+      /* Last buffer that isn't *scratch*. */
+      if (get_buffer_next (bp) && !get_buffer_next (get_buffer_next (bp)))
+        last_bp = bp;
+      c++;
+    }
+  if (c == 3)
+    { /* *scratch* and two files. */
+      FUNCALL (split_window);
+      switch_to_buffer (last_bp);
+      FUNCALL (other_window);
+    }
+  else if (c > 3) /* More than two files. */
+    FUNCALL (list_buffers);
 
   /* Reinitialise the scratch buffer to catch settings */
   init_buffer (scratch_bp);
