@@ -29,28 +29,27 @@
 void
 resync_redisplay (Window * wp)
 {
-  int delta = get_buffer_pt (get_window_bp (wp)).n - get_window_lastpointn (wp);
+  size_t n = offset_to_line (get_window_bp (wp), get_buffer_pt (get_window_bp (wp)));
+  ptrdiff_t delta = n - get_window_lastpointn (wp);
 
   if (delta)
     {
       if ((delta > 0 && get_window_topdelta (wp) + delta < get_window_eheight (wp)) ||
           (delta < 0 && get_window_topdelta (wp) >= (size_t) (-delta)))
         set_window_topdelta (wp, get_window_topdelta (wp) + delta);
-      else if (get_buffer_pt (get_window_bp (wp)).n > get_window_eheight (wp) / 2)
+      else if (n > get_window_eheight (wp) / 2)
         set_window_topdelta (wp, get_window_eheight (wp) / 2);
       else
-        set_window_topdelta (wp, get_buffer_pt (get_window_bp (wp)).n);
+        set_window_topdelta (wp, n);
     }
-  set_window_lastpointn (wp, get_buffer_pt (get_window_bp (wp)).n);
+  set_window_lastpointn (wp, n);
 }
 
 void
 resize_windows (void)
 {
-  Window *wp;
-  int hdelta;
-
   /* Resize windows horizontally. */
+  Window *wp;
   for (wp = head_wp; wp != NULL; wp = get_window_next (wp))
     {
       set_window_fwidth (wp, term_width ());
@@ -59,6 +58,7 @@ resize_windows (void)
 
   /* Work out difference in window height; windows may be taller than
      terminal if the terminal was very short. */
+  int hdelta;
   for (hdelta = term_height () - 1, wp = head_wp;
        wp != NULL; hdelta -= get_window_fheight (wp), wp = get_window_next (wp));
 
@@ -76,8 +76,7 @@ resize_windows (void)
     }
   else
     { /* Decrease windows' height, and close windows if necessary. */
-      int decreased = true;
-      while (decreased)
+      for (int decreased = true; decreased;)
         {
           decreased = false;
           for (wp = head_wp; wp != NULL && hdelta < 0; wp = get_window_next (wp))
@@ -106,12 +105,12 @@ resize_windows (void)
 void
 recenter (Window * wp)
 {
-  Point pt = window_pt (wp);
+  size_t n = offset_to_line (get_window_bp (wp), window_o (wp));
 
-  if (pt.n > get_window_eheight (wp) / 2)
+  if (n > get_window_eheight (wp) / 2)
     set_window_topdelta (wp, get_window_eheight (wp) / 2);
   else
-    set_window_topdelta (wp, pt.n);
+    set_window_topdelta (wp, n);
 }
 
 DEFUN ("recenter", recenter)
@@ -121,6 +120,7 @@ The desired position of point is always relative to the current window.
 +*/
 {
   recenter (cur_wp);
-  term_full_redisplay ();
+  term_clear ();
+  term_redisplay ();
 }
 END_DEFUN
