@@ -34,9 +34,8 @@ static bool
 no_upper (const char *s, size_t len, int regex)
 {
   int quote_flag = 0;
-  size_t i;
 
-  for (i = 0; i < len; i++)
+  for (size_t i = 0; i < len; i++)
     {
       if (regex && s[i] == '\\')
         quote_flag = !quote_flag;
@@ -50,7 +49,7 @@ no_upper (const char *s, size_t len, int regex)
 static const char *re_find_err = NULL;
 
 static int
-find_substr (astr as, const char *s2, size_t s2size, size_t from, size_t to,
+find_substr (castr as1, castr as2, const char *n, size_t nsize, size_t from, size_t to,
              bool forward, bool notbol, bool noteol, bool regex, bool icase)
 {
   int ret = -1;
@@ -67,97 +66,70 @@ find_substr (astr as, const char *s2, size_t s2size, size_t from, size_t to,
   re_set_syntax (syntax);
   search_regs.num_regs = 1;
 
-  re_find_err = re_compile_pattern (s2, (int) s2size, &pattern);
+  re_find_err = re_compile_pattern (n, (int) nsize, &pattern);
   pattern.not_bol = notbol;
   pattern.not_eol = noteol;
   if (!re_find_err)
-    ret = re_search (&pattern, astr_cstr (as), (int) astr_len (as), forward ? from : to - 1,
-                     forward ? (to - from) : -(to - 1 - from), &search_regs);
+    ret = re_search_2 (&pattern,
+                       astr_cstr (as1), (int) astr_len (as1),
+                       astr_cstr (as2), (int) astr_len (as2),
+                       forward ? from : to - 1,
+                       forward ? (to - from) : -(to - 1 - from),
+                       &search_regs,
+                       (int) astr_len (as1) + astr_len (as2));
 
   if (ret >= 0)
-    {
-      ret = forward ? search_regs.end[0] : ret;
-      free (search_regs.start);
-      free (search_regs.end);
-    }
+    ret = forward ? search_regs.end[0] : ret;
 
-  regfree (&pattern);
   return ret;
 }
 
 static bool
-search (Point pt, const char *s, int forward, int regexp)
+search (size_t o, const char *s, int forward, int regexp)
 {
-  Line *lp = pt.p;
-  astr as = get_line_text (lp);
-  size_t ssize = strlen (s), from = 0, to = astr_len (as);
-  bool downcase = get_variable_bool ("case-fold-search") && no_upper (s, ssize, regexp);
-  bool notbol = false, noteol = false;
-  int pos;
-
+  size_t ssize = strlen (s);
   if (ssize < 1)
     return false;
 
-  /* Match first line. */
-  if (forward)
-    {
-      notbol = pt.o > from;
-      from = pt.o;
-    }
-  else
-    {
-      noteol = pt.o < to;
-      to = pt.o;
-    }
-  pos = find_substr (as, s, ssize, from, to, forward, notbol, noteol, regexp, downcase);
-
-  /* Match following lines. */
-  while (pos < 0)
-    {
-      lp = (forward ? get_line_next : get_line_prev) (lp);
-      if (lp == get_buffer_lines (cur_bp))
-        break;
-      as = get_line_text (lp);
-      pos = find_substr (as, s, ssize, 0, astr_len (as), forward, false, false, regexp, downcase);
-    }
-
+  /* Attempt match. */
+  bool notbol = forward ? o > 0 : false;
+  bool noteol = forward ? false : o < get_buffer_size (cur_bp);
+  size_t from = forward ? o : 0;
+  size_t to = forward ? get_buffer_size (cur_bp) : o;
+  int pos = find_substr (get_buffer_pre_point (cur_bp), get_buffer_post_point (cur_bp),
+                         s, ssize, from, to, forward, notbol, noteol, regexp,
+                         get_variable_bool ("case-fold-search") && no_upper (s, ssize, regexp));
   if (pos < 0)
     return false;
 
-  while (get_buffer_pt (cur_bp).p != lp)
-    (forward ? next_line : previous_line) ();
-  pt = get_buffer_pt (cur_bp);
-  pt.o = pos;
-  set_buffer_pt (cur_bp, pt);
+  goto_offset (pos);
   thisflag |= FLAG_NEED_RESYNC;
   return true;
 }
 
-static char *last_search = NULL;
+static castr last_search = NULL;
 
 static le *
-do_search (bool forward, bool regexp, const char *pattern)
+do_search (bool forward, bool regexp, castr pattern)
 {
   le * ok = leNIL;
-  char *ms = NULL;
 
   if (pattern == NULL)
-    pattern = ms = minibuf_read ("%s%s: ", last_search, regexp ? "RE search" : "Search", forward ? "" : " backward");
+    pattern = minibuf_read ("%s%s: ", astr_cstr (last_search),
+                            regexp ? "RE search" : "Search", forward ? "" : " backward");
 
   if (pattern == NULL)
     return FUNCALL (keyboard_quit);
-  if (pattern[0] != '\0')
+  if (astr_len (pattern) != 0)
     {
-      free (last_search);
-      last_search = xstrdup (pattern);
+      last_search = pattern;
 
-      if (!search (get_buffer_pt (cur_bp), pattern, forward, regexp))
+      if (!search (get_buffer_pt (cur_bp), astr_cstr (pattern), forward, regexp))
         minibuf_error ("Search failed: \"%s\"", pattern);
       else
         ok = leT;
     }
 
-  free (ms);
   return ok;
 }
 
@@ -169,7 +141,6 @@ Search forward from point for the user specified text.
 {
   STR_INIT (pattern);
   ok = do_search (true, false, pattern);
-  STR_FREE (pattern);
 }
 END_DEFUN
 
@@ -181,7 +152,6 @@ Search backward from point for the user specified text.
 {
   STR_INIT (pattern);
   ok = do_search (false, false, pattern);
-  STR_FREE (pattern);
 }
 END_DEFUN
 
@@ -193,7 +163,6 @@ Search forward from point for regular expression REGEXP.
 {
   STR_INIT (pattern);
   ok = do_search (true, true, pattern);
-  STR_FREE (pattern);
 }
 END_DEFUN
 
@@ -205,7 +174,6 @@ Search backward from point for match for regular expression REGEXP.
 {
   STR_INIT (pattern);
   ok = do_search (false, true, pattern);
-  STR_FREE (pattern);
 }
 END_DEFUN
 
@@ -215,28 +183,22 @@ END_DEFUN
 static le *
 isearch (int forward, int regexp)
 {
-  int c;
-  int last = true;
-  astr buf = astr_new ();
-  astr pattern = astr_new ();
-  Point start, cur;
   Marker *old_mark = copy_marker (get_buffer_mark (get_window_bp (cur_wp)));
 
-  start = cur = get_buffer_pt (cur_bp);
-
-  /* I-search mode. */
   set_buffer_isearch (get_window_bp (cur_wp), true);
 
+  int last = true;
+  astr pattern = astr_new ();
+  size_t start = get_buffer_pt (cur_bp), cur = start;
   for (;;)
     {
       /* Make the minibuf message. */
-      astr_truncate (buf, 0);
-      astr_afmt (buf, "%sI-search%s: %s",
-                 (last ?
-                  (regexp ? "Regexp " : "") :
-                  (regexp ? "Failing regexp " : "Failing ")),
-                 forward ? "" : " backward",
-                 astr_cstr (pattern));
+      astr buf = astr_fmt ("%sI-search%s: %s",
+                           (last ?
+                            (regexp ? "Regexp " : "") :
+                            (regexp ? "Failing regexp " : "Failing ")),
+                           forward ? "" : " backward",
+                           astr_cstr (pattern));
 
       /* Regex error. */
       if (re_find_err)
@@ -247,17 +209,17 @@ isearch (int forward, int regexp)
             {
               re_find_err = "incomplete input";
             }
-          astr_afmt (buf, " [%s]", re_find_err);
+          astr_cat (buf, astr_fmt (" [%s]", re_find_err));
           re_find_err = NULL;
         }
 
       minibuf_write ("%s", astr_cstr (buf));
 
-      c = getkey ();
+      int c = getkey (GETKEY_DEFAULT);
 
       if (c == KBD_CANCEL)
         {
-          set_buffer_pt (cur_bp, start);
+          goto_offset (start);
           thisflag |= FLAG_NEED_RESYNC;
 
           /* Quit. */
@@ -265,7 +227,7 @@ isearch (int forward, int regexp)
 
           /* Restore old mark position. */
           if (get_buffer_mark (cur_bp))
-            free_marker (get_buffer_mark (cur_bp));
+            unchain_marker (get_buffer_mark (cur_bp));
 
           set_buffer_mark (cur_bp, copy_marker (old_mark));
           break;
@@ -276,7 +238,7 @@ isearch (int forward, int regexp)
             {
               astr_truncate (pattern, astr_len (pattern) - 1);
               cur = start;
-              set_buffer_pt (cur_bp, start);
+              goto_offset (start);
               thisflag |= FLAG_NEED_RESYNC;
             }
           else
@@ -285,7 +247,7 @@ isearch (int forward, int regexp)
       else if (c & KBD_CTRL && (c & 0xff) == 'q')
         {
           minibuf_write ("%s^Q-", astr_cstr (buf));
-          astr_cat_char (pattern, xgetkey (GETKEY_UNFILTERED, 0));
+          astr_cat_char (pattern, getkey_unfiltered (GETKEY_DEFAULT));
         }
       else if (c & KBD_CTRL && ((c & 0xff) == 'r' || (c & 0xff) == 's'))
         {
@@ -299,11 +261,10 @@ isearch (int forward, int regexp)
               /* Find next match. */
               cur = get_buffer_pt (cur_bp);
               /* Save search string. */
-              free (last_search);
-              last_search = xstrdup (astr_cstr (pattern));
+              last_search = astr_cpy (astr_new (), pattern);
             }
           else if (last_search != NULL)
-            astr_cpy_cstr (pattern, last_search);
+            astr_cpy (pattern, last_search);
         }
       else if (c & KBD_META || c & KBD_CTRL || c > KBD_TAB)
         {
@@ -315,11 +276,10 @@ isearch (int forward, int regexp)
                 {
                   /* Save mark. */
                   set_mark ();
-                  set_marker_pt (get_buffer_mark (cur_bp), start);
+                  set_marker_o (get_buffer_mark (cur_bp), start);
 
                   /* Save search string. */
-                  free (last_search);
-                  last_search = xstrdup (astr_cstr (pattern));
+                  last_search = astr_cpy (astr_new (), pattern);
 
                   minibuf_write ("Mark saved when search started");
                 }
@@ -345,11 +305,8 @@ isearch (int forward, int regexp)
   /* done */
   set_buffer_isearch (get_window_bp (cur_wp), false);
 
-  astr_delete (buf);
-  astr_delete (pattern);
-
   if (old_mark)
-    free_marker (old_mark);
+    unchain_marker (old_mark);
 
   return leT;
 }
@@ -406,6 +363,25 @@ is treated as a regexp.  See @kbd{M-x isearch-forward} for more info.
 }
 END_DEFUN
 
+/*
+ * Check the case of a string.
+ * Returns 2 if it is all upper case, 1 if just the first letter is,
+ * and 0 otherwise.
+ */
+static int
+check_case (astr as)
+{
+  size_t i;
+  for (i = 0; i < astr_len (as) && isupper ((int) astr_get (as, i)); i++)
+    ;
+  if (i == astr_len (as))
+    return 2;
+  else if (i == 1)
+    for (; i < astr_len (as) && !isupper ((int) astr_get (as, i)); i++)
+      ;
+  return i == astr_len (as);
+}
+
 DEFUN ("query-replace", query_replace)
 /*+
 Replace occurrences of a string with other text.
@@ -413,31 +389,21 @@ As each match is found, the user must type a character saying
 what to do with it.
 +*/
 {
-  bool noask = false, find_no_upper;
-  size_t find_len, count = 0;
-  char *find = minibuf_read ("Query replace string: ", "");
-  char *repl;
-
+  castr find = minibuf_read ("Query replace string: ", "");
   if (find == NULL)
     return FUNCALL (keyboard_quit);
-  if (*find == '\0')
-    {
-      free (find);
-      return leNIL;
-    }
-  find_len = strlen (find);
-  find_no_upper = no_upper (find, find_len, false);
+  if (astr_len (find) == 0)
+    return leNIL;
+  bool find_no_upper = no_upper (astr_cstr (find), astr_len (find), false);
 
-  repl = minibuf_read ("Query replace `%s' with: ", "", find);
+  castr repl = minibuf_read ("Query replace `%s' with: ", "", find);
   if (repl == NULL)
-    {
-      free (find);
-      return FUNCALL (keyboard_quit);
-    }
+    return FUNCALL (keyboard_quit);
 
-  while (search (get_buffer_pt (cur_bp), find, true, false))
+  bool noask = false;
+  size_t count = 0;
+  while (search (get_buffer_pt (cur_bp), astr_cstr (find), true, false))
     {
-      Point pt;
       int c = ' ';
 
       if (!noask)
@@ -448,13 +414,13 @@ what to do with it.
             {
               minibuf_write
                 ("Query replacing `%s' with `%s' (y, n, !, ., q)? ", find,
-                 repl);
-              c = getkey ();
+                 astr_cstr (repl));
+              c = getkey (GETKEY_DEFAULT);
               if (c == KBD_CANCEL || c == KBD_RET || c == ' ' || c == 'y'
                   || c == 'n' || c == 'q' || c == '.' || c == '!')
                 break;
               minibuf_error ("Please answer y, n, !, . or q.");
-              waitkey (WAITKEY_DEFAULT);
+              waitkey ();
             }
           minibuf_clear ();
 
@@ -472,19 +438,28 @@ what to do with it.
         }
 
       /* Perform replacement. */
-      pt = get_buffer_pt (cur_bp);
       ++count;
-      undo_save (UNDO_REPLACE_BLOCK,
-                 make_point (pt.n, pt.o - find_len), find_len, strlen (repl));
-      line_replace_text (pt.p, pt.o - find_len, find_len, repl, find_no_upper);
+      castr case_repl = repl;
+      Region r = region_new (get_buffer_pt (cur_bp) - astr_len (find), get_buffer_pt (cur_bp));
+      if (find_no_upper && get_variable_bool ("case-replace"))
+        {
+          int case_type = check_case (get_buffer_region (cur_bp, r).as);
+
+          if (case_type != 0)
+            case_repl = astr_recase (astr_cpy (astr_new (), repl),
+                                     case_type == 1 ? case_capitalized : case_upper);
+        }
+
+      Marker *m = point_marker ();
+      goto_offset (r.start);
+      replace_estr (astr_len (find), estr_new_astr (case_repl));
+      goto_offset (get_marker_o (m));
+      unchain_marker (m);
 
       if (c == '.')		/* Replace and quit. */
         break;
     }
 
-  free (find);
-  free (repl);
-
   if (thisflag & FLAG_NEED_RESYNC)
     resync_redisplay (cur_wp);
 
