@@ -65,15 +65,20 @@ term_clear (void)
 }
 
 void
-term_addch (int c)
+term_addstr (const char *s)
 {
-  addch ((chtype) (c & ~A_ATTRIBUTES));
+  addstr (s);
 }
 
 void
 term_attrset (size_t attr)
 {
-  attrset (attr == FONT_REVERSE ? A_REVERSE : 0);
+  int attrs = 0;
+  if (attr & FONT_REVERSE)
+    attrs |= A_REVERSE;
+  if (attr & FONT_UNDERLINE)
+    attrs |= A_UNDERLINE;
+  attrset (attrs);
 }
 
 void
@@ -82,11 +87,22 @@ term_beep (void)
   beep ();
 }
 
+size_t
+term_width (void)
+{
+  return (size_t) COLS;
+}
+
+size_t
+term_height (void)
+{
+  return (size_t) LINES;
+}
+
 void
 term_init (void)
 {
   initscr ();
-  term_set_size ((size_t) COLS, (size_t) LINES);
   noecho ();
   nonl ();
   raw ();
@@ -99,8 +115,7 @@ term_init (void)
 void
 term_close (void)
 {
-  /* Free memory and finish with ncurses. */
-  gl_list_free (key_buf);
+  /* Finish with ncurses. */
   endwin ();
 }
 
@@ -205,13 +220,11 @@ codetokey (int c)
 static size_t
 keytocodes (size_t key, int ** codevec)
 {
-  int * p;
-
-  p = *codevec = XCALLOC (2, int);
-
   if (key == KBD_NOKEY)
     return 0;
 
+  int *p = *codevec = XCALLOC (2, int);
+
   if (key & KBD_META)				/* META */
     *p++ = '\33';
   key &= ~KBD_META;
@@ -335,58 +348,52 @@ keytocodes (size_t key, int ** codevec)
 }
 
 static int
-get_char (void)
+get_char (int delay)
 {
   int c;
-  size_t size = term_buf_len ();
 
+  size_t size = term_buf_len ();
   if (size > 0)
     {
       c = (ptrdiff_t) gl_list_get_at (key_buf, size - 1);
       gl_list_remove_at (key_buf, size - 1);
     }
   else
-    c = getch ();
+    {
+      timeout (delay);
+
+#ifdef KEY_RESIZE
+      do {
+#endif
+        c = getch ();
+
+#ifdef KEY_RESIZE
+        if (c == KEY_RESIZE)
+          resize_windows ();
+      } while (c == KEY_RESIZE);
+#endif
+
+      timeout (-1);
+    }
 
   return c;
 }
 
 size_t
-term_xgetkey (int mode, size_t delay)
+term_getkey (int delay)
 {
-  size_t key;
-
-  for (;;)
-    {
-      int c;
-
-      if (mode & GETKEY_DELAYED)
-        wtimeout (stdscr, (int) delay * 100);
-
-      c = get_char ();
-      if (mode & GETKEY_DELAYED)
-        wtimeout (stdscr, -1);
-
-#ifdef KEY_RESIZE
-      if (c == KEY_RESIZE)
-        {
-          term_set_size ((size_t) COLS, (size_t) LINES);
-          resize_windows ();
-          continue;
-        }
-#endif
-
-      if (mode & GETKEY_UNFILTERED)
-        key = (size_t) c;
-      else
-        {
-          key = codetokey (c);
-          while (key == KBD_META)
-            key = codetokey (get_char ()) | KBD_META;
-        }
-      break;
-    }
+  size_t key = codetokey (get_char (delay));
+  while (key == KBD_META)
+    key = codetokey (get_char (GETKEY_DEFAULT)) | KBD_META;
+  return key;
+}
 
+int
+term_getkey_unfiltered (int delay)
+{
+  keypad (stdscr, false);
+  int key = get_char (delay);
+  keypad (stdscr, true);
   return key;
 }
 
@@ -394,10 +401,6 @@ void
 term_ungetkey (size_t key)
 {
   int * codes = NULL;
-  size_t i, n = keytocodes (key, &codes);
-
-  for (i = n; i > 0; i--)
+  for (size_t i = keytocodes (key, &codes); i > 0; i--)
     gl_list_add_last (key_buf, (void *)(ptrdiff_t) codes[i - 1]);
-
-  free (codes);
 }
